from yta_editor_frame.frame.metadata import _FrameMetadata, _FrameMetadataExtractor
from yta_editor_frame.frame.resource.types import FrameResourceType
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union


class Frame:
    """
    A frame, the main class that is shared in between
    the nodes and all the rendering process.

    If the `metadata` is provided, it will be set. If
    it is not provided, it will be extracted if the
    `do_autoextract_metadata` is `True`, or set as
    unknown if `False`.
    """

    def __init__(
        self,
        # TODO: Set the 4 types we have
        resource,
        render_context: 'RenderContext',
        metadata: Union[dict, None] = None,
        do_autoextract_metadata: bool = False
    ):
        ParameterValidator.validate_dict('metadata', metadata)

        # TODO: Set the 4 types we have
        self.resource = resource
        self._render_context: 'RenderContext' = render_context
        """
        *For internal use only*

        The `RenderContext` instance that allows us to use
        the `RenderTargetProvider` and more resource 
        handlers.
        """
        self._ref_count: int = 1
        """
        *For internal use only*

        Internal counter to determine when no one is using
        this frame and we can destroy it and free the
        resources.
        """

        if metadata is None:
            self._reset_metadata()
            if do_autoextract_metadata:
                self._autodetect_metadata()
        else:
            metadata = _FrameMetadata.from_dict(
                frame = self,
                metadata = metadata
            )

        self.metadata: _FrameMetadata = metadata
        """
        Metadata we want to include about the frame.
        """

    @classmethod
    def from_frame(
        cls,
        frame: Union['Frame', 'av.VideoFrame', 'np.ndarray', 'moderngl.Texture', 'RenderTarget'],
        render_context: 'RenderContext',
        metadata: Union[dict, None] = None
    ) -> 'Frame':
        """
        Get a `Frame` instance by creating the resource from
        the `frame` parameter provided, that must be one of
        the following:
        - `av.VideoFrame`
        - `np.ndarray`
        - `moderngl.Texture`
        - `RenderTarget`
        """
        # TODO: Refactor if possible (?)
        from yta_editor_frame.frame.resource.av_video_frame_resource import AvVideoFrameResource
        from yta_editor_frame.frame.resource.cpu_frame_resource import CPUNumpyArrayFrameResource
        from yta_editor_frame.frame.resource.gpu_frame_resource import GPUTextureFrameResource
        from yta_editor_frame.frame.resource.render_target_frame_resource import RenderTargetFrameResource

        ParameterValidator.validate_mandatory_instance_of('frame', frame, ['Frame', 'VideoFrame', 'ndarray', 'Texture', 'RenderTarget'])
        ParameterValidator.validate_mandatory_instance_of('render_context', render_context, 'RenderContext')
        ParameterValidator.validate_dict('metadata', metadata)

        if PythonValidator.is_instance_of(frame, 'Frame'):
            return frame

        resource = (
            AvVideoFrameResource(frame)
            if PythonValidator.is_instance_of(frame, 'VideoFrame') else
            CPUNumpyArrayFrameResource(frame)
            if PythonValidator.is_instance_of(frame, 'ndarray') else
            GPUTextureFrameResource(frame)
            if PythonValidator.is_instance_of(frame, 'Texture') else
            RenderTargetFrameResource(frame)
            # if PythonValidator.is_instance_of('RenderTarget') else
        )

        return Frame(
            resource = resource,
            render_context = render_context,
            metadata = metadata
            # TODO: Maybe send 'do_autoextract_metadata' as True (?)
        )

    def _reset_metadata(
        self
    ) -> 'Frame':
        """
        *For internal use only*

        Reset the metadata to hold nothing specific.
        
        This method will remove any additional dynamic
        metadata added before.
        """
        self.metadata = _FrameMetadata(
            frame = self
        )

        return self
    
    def _autodetect_metadata(
        self
    ) -> 'Frame':
        """
        *For internal use only*

        Autodetect the metadata from the resource.

        This method will remove any additional dynamic
        metadata added before.
        """
        self.metadata = _FrameMetadataExtractor.extract(self.resource.frame)

        return self

    def _to(
        self,
        type: FrameResourceType
    ) -> 'Frame':
        """
        *For internal use only*

        Get a new instance with the same metadata but
        of the `type` provided, that must be one of
        the following:
        - `FrameResourceType.GPU`
        - `FrameResourceType.CPU`
        - `FrameResourceType.CPU_AV_VIDEOFRAME`
        """
        from yta_editor_frame.frame.resource.converter import FrameResourceConverter

        type = FrameResourceType.to_enum(type)

        if type not in [FrameResourceType.CPU, FrameResourceType.GPU, FrameResourceType.CPU_AV_VIDEOFRAME]:
            raise Exception(f'The "type" {type.value} is not supported for the transformation.')

        new_resource = (
            FrameResourceConverter.to_numpy(
                resource = self.resource
            )
            if type == FrameResourceType.CPU else
            FrameResourceConverter.to_texture(
                resource = self.resource,
                moderngl_context = self._render_context.opengl_context
            )
            if type == FrameResourceType.GPU else
            FrameResourceConverter.to_videoframe(
                resource = self.resource
            )
        )

        """
        By now we are returning a new instance to
        avoid errors if this `Frame` instance is
        shared and being used by other instances.
        """
        frame = Frame(
            resource = new_resource,
            render_context = self._render_context,
            metadata = None
        )
        frame.metadata = self.metadata.copy(frame)

        return frame
    
    def to_texture(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.GPU`, that is the
        one that works with `moderngl.Texture`.
        """
        return self._to(type = FrameResourceType.GPU)
    
    def to_numpy(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.CPU`, that is the
        one that works with `np.ndarray`.
        """
        return self._to(type = FrameResourceType.CPU)
    
    def to_videoframe(
        self
    ) -> 'Frame':
        """
        Get a new instance with the same metadata but
        of the `FrameResourceType.CPU_AV_VIDEOFRAME`,
        that is the one that works with `av.VideoFrame`.
        """
        return self._to(type = FrameResourceType.CPU_AV_VIDEOFRAME)

    def _as(
        self,
        type: FrameResourceType
    ) -> 'Frame':
        """
        *For internal use only*

        Modify the current instance to transform it
        into one with the `type` provided, that must
        be one of the following:
        - `FrameResourceType.GPU`
        - `FrameResourceType.CPU`
        - `FrameResourceType.CPU_AV_VIDEOFRAME`
        """
        from yta_editor_frame.frame.resource.converter import FrameResourceConverter

        type = FrameResourceType.to_enum(type)

        if type not in [FrameResourceType.CPU, FrameResourceType.GPU, FrameResourceType.CPU_AV_VIDEOFRAME]:
            raise Exception(f'The "type" {type.value} is not supported for the transformation.')
        
        if self._ref_count > 1:
            raise Exception('This "Frame" instance is being used by more than one instance.')
        
        old_resource = self.resource

        new_resource = (
            FrameResourceConverter.to_numpy(
                resource = self.resource
            )
            if type == FrameResourceType.CPU else
            FrameResourceConverter.to_texture(
                resource = self.resource,
                moderngl_context = self._render_context.render_resources.opengl_context
            )
            if type == FrameResourceType.GPU else
            FrameResourceConverter.to_videoframe(
                resource = self.resource
            )
        )

        self.resource = new_resource

        # Release the old resource
        old_resource.release(self._render_context)

        # TODO: What about the metadata? It should be
        # updated somehow, no (?)
        self.metadata = (
            self.metadata.copy()
            if self.metadata is not None else
            None
        )

        return self
    
    def as_numpy(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.CPU` that is the one that
        works with the `np.ndarray`.
        """
        return self._as(FrameResourceType.CPU)
    
    def as_texture(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.GPU` that is the one that
        works with the `moderngl.Texture`.
        """
        return self._as(FrameResourceType.GPU)
    
    def as_videoframe(
        self
    ) -> 'Frame':
        """
        Modify the current instance to transform it into
        a `FrameResourceType.CPU_AV_VIDEOFRAME` that is
        the one that works with the `av.VideoFrame`.
        """
        return self._as(FrameResourceType.CPU_AV_VIDEOFRAME)
    
    def retain(
        self
    ) -> 'Frame':
        """
        Method to indicate that the frame is being used by
        someone else so it must be retained and not sent
        back to the pool.
        """
        if self._ref_count <= 0:
            raise Exception('Cannot retain destroyed frame.')

        self._ref_count += 1

        return self

    def release(
        self
    ):
        """
        Method to indicate that the frame has been freed
        by one owner and could be sent back to the pool
        if no one else is using it.

        If no one is owning this `Frame` instance, it
        will be destroyed, and if the instance is a
        `RenderTarget` instance, it will be sent back to
        the pool.
        """
        self._ref_count -= 1

        if self._ref_count < 0:
            raise Exception('Frame released too many times.')

        if self._ref_count == 0:
            self._destroy()

    def _destroy(
        self
    ):
        """
        Method to destroy the frame instance, that will
        send the resource to the pool if it is a
        `RenderTarget` instance.
        """
        self.resource.release(self._render_context)
        self.resource = None

    def __repr__(
        self
    ):
        return f'<Frame rc={self._ref_count} res={self.resource}>'


from yta_editor_frame.frame.resource.abstract import FrameResourceABC
from yta_editor_frame.frame.resource.types import FrameResourceType


class AvVideoFrameResource(
    FrameResourceABC
):
    """
    A frame resource based on a pyav VideoFrame.

    The `.frame` attribute is an `av.VideoFrame`.
    """

    @property
    def width(
        self
    ) -> int:
        return self.frame.width
    
    @property
    def height(
        self
    ) -> int:
        return self.frame.height
    
    @property
    def format(
        self
    ) -> str:
        return str(self.frame.format.name)
    
    @property
    def type(
        self
    ) -> FrameResourceType:
        return FrameResourceType.CPU_AV_VIDEOFRAME

    def __init__(
        self,
        frame: 'av.VideoFrame'
    ):
        self.frame: 'av.VideoFrame' = frame
        """
        The `frame` itself, as an `av.VideoFrame`.
        """

    def release(
        self,
        render_context: 'RenderContext'
    ):
        pass


from yta_constants.enum import YTAEnum as Enum


class FrameResourceType(Enum):
    """
    The different frame resource types we have.
    """

    CPU = 'cpu'
    """
    The basic CPU frame resource we use to make
    transformations with the nodes and to work
    with:
    - `np.ndarray`
    """
    GPU = 'gpu'
    """
    The basic GPU frame resource we use to make
    transformations when we need to upload 
    something from CPU to GPU or similar:
    - `moderngl.Texture`
    """
    CPU_AV_VIDEOFRAME = 'cpu_av_videoframe'
    """
    The CPU frame resource that we obtain when
    reading a video or that we use to write into
    a new video file:
    - `pyav.VideoFrame`
    """
    GPU_RENDER_TARGET = 'gpu_render_target'
    """
    The GPU frame resource that we manage in our
    own way (with our own created class) to be
    able to write properly with the nodes we
    have and being efficient with the resources
    and memory:
    - `render_target`
    """


from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from dataclasses import dataclass
from typing import Union


@dataclass
class _FrameMetadata:
    """
    *For internal use only*

    Class to wrap the metadata about a frame. Any value
    that is `None` means that we don't have any data
    about it yet.

    The metadata is useful to know if the frame has any
    kind of transparency, if it is fully transparent or
    fully opaque, etc.
    """

    @property
    def width(
        self
    ) -> int:
        """
        The width of the frame, obtained from the resource.
        """
        return self._frame.resource.width
    
    @property
    def height(
        self
    ) -> int:
        """
        The height of the frame, obtained from the resource.
        """
        return self._frame.resource.height
    
    @property
    def format(
        self
    ) -> str:
        """
        The format of the frame, obtained from the resource.
        """
        return self._frame.resource.format

    def __init__(
        self,
        frame: 'Frame',
        has_alpha_pixels: Union[bool, None] = None,
        is_fully_opaque: Union[bool, None] = None,
        is_fully_transparent: Union[bool, None] = None,
        **kwargs
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, 'Frame')

        self._frame: 'Frame' = frame
        """
        *For internal use only*

        The `Frame` instance this metadata belongs to.
        """
        self.has_alpha_pixels: Union[bool, None] = has_alpha_pixels
        """
        Boolean flag to indicate if there is at least one
        pixel with some kind of transparency or not.
        """
        self.is_fully_opaque: Union[bool, None] = is_fully_opaque
        """
        Boolean flag to indicate if the frame is completely
        opaque and there are no transparencies on it. This
        is very useful when we need to combine frames or
        we need to apply an effect based on the transparency,
        so we can process it faster as we know no frame after
        this one will be seen.
        """
        self.is_fully_transparent: Union[bool, None] = is_fully_transparent
        """
        Boolean flag to indicate if the frame is completely
        transparent. This is very useful when we need to
        combine frames because we can skip it when combining.
        """

        # Set the dynamic values
        for key, value in kwargs.items():
            setattr(self, key, value)

    def copy(
        self,
        frame: 'Frame'
    ) -> '_FrameMetadata':
        """
        Get a copy of this `_FrameMetadata` instance but
        attached to the `frame` provided.
        """
        return _FrameMetadata.from_dict(
            frame = frame,
            metadata = {
                'has_alpha_pixels': self.has_alpha_pixels,
                'is_fully_opaque': self.is_fully_opaque,
                'is_fully_transparent': self.is_fully_transparent
                # TODO: What about the other possible fields (?)
            }
        )

    @classmethod
    def from_dict(
        cls,
        frame: 'Frame',
        metadata: dict
    ) -> '_FrameMetadata':
        """
        Get a `_FrameMetadata` instance associated to the
        `frame` provided with the data included in the
        `metadata` dict given.

        This method is useful to build and attach metadata
        when using the extractor.
        """
        # Dynamically add mandatory keys to metadata
        # if missing
        keys = {
            'has_alpha_pixels',
            'is_fully_opaque',
            'is_fully_transparent'
        }

        for key in keys:
            if key not in metadata:
                metadata[key] = None

        return _FrameMetadata(
            frame = frame,
            **metadata
            # has_alpha_pixels = has_alpha_pixels,
            # is_fully_opaque = is_fully_opaque,
            # is_fully_transparent = is_fully_transparent
        )
        

class _FrameMetadataExtractor:
    """
    Class to extract the information of a frame
    to a `_FrameMetadata` instance.
    """

    @staticmethod
    def extract(
        frame: Union['np.ndarray', 'moderngl.Texture', 'av.VideoFrame', 'RenderTarget']
    ) -> _FrameMetadata:
        """
        Extract the metadata from the given `frame`
        that is easy to get and available.
        
        There is metadata that has to be manually
        written when the frame is modified by a node
        and we know exactly some metadata of the
        output we will generate and return, but this
        is just an easy and automated extractor.
        """
        metadata_dict = {
            'has_alpha_pixels': None,
            'is_fully_opaque': None,
            'is_fully_transparent': None
        }

        # TODO: Extract it
        if PythonValidator.is_numpy_array(frame):
            # [1] `np.ndarray`
            if (
                frame.ndim == 3 and
                frame.shape[2] == 4
            ):
                alpha = frame[..., 3]
                a_min = alpha.min()
                a_max = alpha.max()
                metadata_dict['has_alpha_pixels'] = a_min < 255
                metadata_dict['is_fully_opaque'] = a_min == 255
                metadata_dict['is_fully_transparent'] = a_max == 0
            else:
                metadata_dict['has_alpha_pixels'] = False
                metadata_dict['is_fully_opaque'] = True
                metadata_dict['is_fully_transparent'] = False
        elif PythonValidator.is_instance_of(frame, 'VideoFrame'):
            # [2] `av.VideoFrame`
            metadata_dict['has_alpha_pixels'] = 'a' in frame.format.name
            metadata_dict['is_fully_opaque'] = (
                True
                if not metadata_dict['has_alpha_pixels'] else
                # We don't know yet
                None
            )
            metadata_dict['is_fully_transparent'] = (
                False
                if not metadata_dict['has_alpha_pixels'] else
                # We don't know yet
                None
            )

        """
        # [3] `moderngl.Texture`
        # [4] `RenderTarget`
        
        The `moderngl.Texture` (or `RenderTarget`, that has
        a texture inside) are not easy to access, so we
        don't. If we are using a specific node that we know
        that will generate no transparency, then we can set
        that metadata manually, but this is to extract the
        information automatically.
        """

        return _FrameMetadata.from_dict(
            frame = frame,
            metadata = metadata_dict
        )


from yta_programming.decorators.requires_dependency import requires_dependency
from yta_editor_utils.texture import TextureUtils
from typing import Union


class FrameHelper:
    """
    Class to simplify the way we handle frames from
    their different sources and according to the
    different and possible types.

    Possible frame types:
    - `av.VideoFrame`
    - `np.ndarray`
    - `moderngl.Texture`
    """

    @staticmethod
    @requires_dependency('moderngl', 'yta_editor_frame', 'moderngl')
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def ndarray_to_texture(
        frame: 'np.ndarray',
        moderngl_context: Union['moderngl.Context', None],
        do_include_alpha: bool = True,
        do_flip: bool = True
    ) -> 'moderngl.Texture':
        """
        *Optional dependency `moderngl` (imported as `moderngl`) required*

        Transform the `numpy.ndarray` `frame` provided into a
        `moderngl.Texture` frame.

        The `do_flip` must be `True` normally because the
        origin of ModernGL and the origin of numpy are
        inverted.
        """
        # TODO: This is creating a new texture everytime,
        # and not just writing the data to a pre-created one
        return TextureUtils.numpy_to_texture(
            input = frame,
            opengl_context = moderngl_context,
            # TODO: Can I detect it (?)
            dtype = 'f1',
            do_flip = True
        )

        # TODO: Below is outdated, remove it soon
        import moderngl
        import numpy as np

        # TODO: Create 'moderngl_context' if None. I have a
        # class called 'OpenGLContext' to get it

        if frame.ndim != 3:
            raise ValueError('Expected HWC ndarray')

        dtype_map = {
            np.float32: 'f4',
            np.uint8: 'u1'
        }
        dtype = dtype_map.get(frame.dtype, 'u1')

        # We consider HWC format to obtain the size
        height, width = frame.shape[:2]

        # TODO: Is this flip here ok (?)
        frame = np.flip(frame, axis = 0)
        # This is needed because we are creating a texture
        frame = np.ascontiguousarray(frame)

        return moderngl_context.texture(
            (width, height),
            # This was originally 4 only but I added the bool
            # components = (
            #     4
            #     if do_include_alpha else
            #     3
            # ),
            # TODO: Apparently this above is not a good thing
            components = frame.shape[2],
            # data = frame.tobytes(),
            # TODO: Apparently 'frame' is accepted directly
            data = frame,
            dtype = dtype
        )
    
    @staticmethod
    @requires_dependency('av', 'yta_editor_frame', 'av')
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def ndarray_to_videoframe(
        frame: 'np.ndarray',
        do_include_alpha: bool = True,
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None
    ) -> 'av.VideoFrame':
        """
        *Optional dependency `av` (imported as `av`) required*

        Transform the `numpy.ndarray` `frame` provided into a
        `av.VideoFrame` frame.
        """
        video_frame = TextureUtils.numpy_frame_to_pyav_frame(
            frame = frame
        )

        if pts is not None:
            video_frame.pts = pts

        if time_base is not None:
            video_frame.time_base = time_base

        return video_frame

        # TODO: Below is outdated, remove it soon
        import av

        format = (
            'rgba'
            if do_include_alpha else
            'rgb24'
        )

        frame = av.VideoFrame.from_ndarray(
            array = frame,
            format = format
        )

        if pts is not None:
            frame.pts = pts

        if time_base is not None:
            frame.time_base = time_base

        return frame

    @staticmethod
    @requires_dependency('av', 'yta_editor_frame', 'av')
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def videoframe_to_ndarray(
        frame: 'av.VideoFrame',
        do_include_alpha: bool = True
    ) -> 'np.ndarray':
        """
        *Optional dependency `av` (imported as `av`) required*

        Transform the `av.VideoFrame` `frame` provided into a
        `numpy.ndarray` frame.
        """
        format = (
            'rgba'
            if do_include_alpha else
            'rgb24'
        )

        return frame.to_ndarray(format = format)
    
    @staticmethod
    @requires_dependency('av', 'yta_editor_frame', 'av')
    @requires_dependency('moderngl', 'yta_editor_frame', 'moderngl')
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def videoframe_to_texture(
        frame: 'av.VideoFrame',
        moderngl_context: 'moderngl.Context',
        do_include_alpha: bool = True,
        do_flip: bool = True
    ) -> 'moderngl.Texture':
        """
        *Optional dependency `av` (imported as `av`) required*

        *Optional dependency `moderngl` (imported as `moderngl`) required*

        Transform the `av.VideoFrame` `frame` provided into a
        `moderngl.Texture` frame.

        The `do_flip` must be `True` normally because the
        origin of ModernGL and the origin of numpy are
        inverted.
        """
        # 1. 'av.VideoFrame' to 'np.ndarray'
        frame = FrameHelper.videoframe_to_ndarray(
            frame = frame,
            do_include_alpha = do_include_alpha
        )

        # 2. 'np.ndarray' to 'moderngl.Texture'
        return FrameHelper.ndarray_to_texture(
            frame = frame,
            moderngl_context = moderngl_context,
            do_include_alpha = do_include_alpha,
            do_flip = do_flip
        )
    
    @staticmethod
    @requires_dependency('moderngl', 'yta_editor_frame', 'moderngl')
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def texture_to_ndarray(
        frame: 'moderngl.Texture',
        do_include_alpha: bool = True,
        do_flip: bool = True,
        do_force_contiguous: bool = False
    ) -> 'np.ndarray':
        """
        *Optional dependency `moderngl` (imported as `moderngl`) required*

        Transform the `moderngl.Texture` `frame` provided into
        a `np.ndarray` frame.

        This method does a `flipud` in the content.

        (!) This method returns a tuple that includes the array
        but also a boolean indicating if the texture included
        alpha or not.

        The `do_flip` must be `True` normally because the
        origin of ModernGL and the origin of numpy are
        inverted.

        The `do_force_contiguous` must be `True` only if the
        numpy will be transformed inmediately into a pyav
        videoframe (that needs it contiguous). If you will
        operate with it, set it to `False`.
        """
        return TextureUtils.texture_to_numpy(
            texture = frame,
            # TODO: Can I adjust this (?)
            do_include_alpha = do_include_alpha,
            do_flip = do_flip,
            do_force_contiguous = do_force_contiguous
        )

        # TODO: Below is outdated, remove it soon
        import numpy as np

        data = frame.read()
        width, height = frame.size

        dtype_map = {
            'f4': np.float32,
            # 'f2': np.float16,
            'f1': np.uint8,
            'u1': np.uint8
        }
        dtype = dtype_map.get(frame.dtype, np.uint8)

        data = np.frombuffer(data, dtype = dtype)

        if frame.dtype == 'f1':
            data = data * 255.0

        data = data.reshape((height, width, frame.components))

        data = np.flipud(data)
        data = np.ascontiguousarray(data)
        # Similar to 'np.flipud(frame)'
        # data = np.flip(data, axis = 0)

        """
        The 'frame.components defines if RGBA or RGB:
        - `frame.components=4` means RGBA
        - `frame.components=3` means RGB
        """
        return (data, frame.components == 4)

    @staticmethod
    @requires_dependency('av', 'yta_editor_frame', 'av')
    @requires_dependency('moderngl', 'yta_editor_frame', 'moderngl')
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def texture_to_videoframe(
        frame: 'moderngl.Texture',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        do_include_alpha: bool = True,
        do_flip: bool = True
    ) -> 'av.VideoFrame':
        """
        *Optional dependency `av` (imported as `av`) required*

        *Optional dependency `moderngl` (imported as `moderngl`) required*

        Transform the `moderngl.Texture` `frame` provided into
        a `av.VideoFrame` frame.

        The `do_flip` must be `True` normally because the
        origin of ModernGL and the origin of numpy are
        inverted.
        """
        video_frame = TextureUtils.texture_to_pyav_frame(
            texture = frame,
            # TODO: Can I adjust this (?)
            do_include_alpha = do_include_alpha,
            do_flip = do_flip,
            do_force_contiguous = True
        )

        if pts is not None:
            video_frame.pts = pts

        if time_base is not None:
            video_frame.time_base = time_base

        return video_frame

        # TODO: Below is outdated, remove it soon
        # 1. 'moderngl.Texture' to 'np.ndarray'
        frame, do_include_alpha = FrameHelper.texture_to_ndarray(
            frame = frame
        )

        # 2. 'np.ndarray' to 'av.VideoFrame'
        frame = FrameHelper.ndarray_to_videoframe(
            frame = frame,
            do_include_alpha = do_include_alpha
        )

        if pts is not None:
            frame.pts = pts

        if time_base is not None:
            frame.time_base = time_base

        return frame


"""
This 'FrameBuffer' concept has been removed 
recently to be replaced by 'Frame'. I keep
the code some commits just in case we have a
bug and I want to check quickly how it was
implemented here to fix it. It will disappear,
don't worry.
"""
# from yta_editor_frame.domain import FrameBufferDomain
# from yta_editor_frame.utils import _texture_to_numpy_frame, _ensure_rgba_numpy_array, _numpy_frame_to_texture
# from yta_constants.enum import YTAEnum as Enum
# from yta_programming.decorators.requires_dependency import requires_dependency
# from yta_validation.parameter import ParameterValidator
# from yta_validation import PythonValidator
# from dataclasses import dataclass, field
# from typing import Union


# # TODO: Move these Enums, please
# class FrameBufferFormat(Enum):
#     """
#     The format of the frame buffer that will identify
#     if we have alpha channel or not.
#     """

#     RGBA = 'rgba'
#     """
#     The `rgba` format.
#     """

# class FrameBufferDType(Enum):
#     """
#     The numpy dtype of the frame buffer that will be
#     helpful to operate with it.
#     """

#     UINT8 = 'uint8'
#     """
#     The `uint8` dtype.
#     """

# FORMAT_TO_FORCE = 'rgba'

# @dataclass
# class FrameBuffer:
#     """
#     *Dataclass*

#     Dataclass to store the information about a buffer
#     that we will need during the processing.

#     This library could need one of these optional
#     dependencies depending on what method you call to
#     use it in the context you need it.
#     - `av`
#     - `numpy`
#     - `moderngl`

#     We will always store the frames as they are 
#     provided, but forced to be 'RGBA' when returned
#     with the `as_...` method.
#     """

#     # Common attributes
#     size: tuple[int, int]
#     """
#     The size of the frame, in pixels, expressed as a
#     `(width, height)` tuple.
#     """

#     """
#     TODO: Maybe we don't need this one, if it is
#     fully opaque we treat it as non-transparent,
#     and if it is not opaque nor transparent, it
#     means that it has some transparency so we have
#     to handle it and this 'has_alpha_pixels' would
#     be automatically yes.
#     """
#     has_alpha_pixels: Union[bool, None] = field(default = None)
#     """
#     Flag to indicate if the frame has alpha pixels.

#     The value could be `None` if we don't actually
#     know if yes or no (maybe because it came from an
#     unknown `moderngl.Texture`).
#     """
#     is_fully_opaque: Union[bool, None] = field(default = None)
#     """
#     Flag to indicate if the frame is fully opaque,
#     which means that the alpha channel is giving no
#     transparency to this frame. This is useful for
#     the blending process because we can stop blending
#     when this frame appears (the next ones will not be
#     shown due to the opacity of this one).

#     The value could be `None` if we don't actually
#     know if yes or no (maybe because it came from an
#     unknown `moderngl.Texture`).
#     """
#     is_fully_transparent: Union[bool, None] = field(default = None)
#     """
#     Flag to indicate if the frame is fully transparent,
#     which means that the alpha channel is giving full
#     transparency to this frame. This is ueful for the
#     blending process because we can omit it when its
#     not the only frame to combine (to blend).

#     The value could be `None` if we don't actually
#     know if yes or no (maybe because it came from an
#     unknown `moderngl.Texture`).
#     """

#     # Attributes if 'av.VideoFrame'
#     @property
#     def av_format(
#         self
#     ) -> Union[str, None]:
#         """
#         The `format` of the frame as a `av.VideoFrame`, if
#         existing.
#         """
#         return (
#             self._av_videoframe.format.name
#             if self._av_videoframe is not None else
#             None
#         )
    
#     @property
#     def av_planes(
#         self
#     ) -> Union[int, None]:
#         """
#         The number of `planes` of the frame as a
#         `av.VideoFrame`, if existing.
#         """
#         return (
#             len(self._av_videoframe.planes)
#             if self._av_videoframe is not None else
#             None
#         )

#     # Attributes if 'np.ndarray'
#     @property
#     def np_dtype(
#         self
#     ) -> Union['np.dtype', None]:
#         """
#         The `dtype` of the frame as a `np.ndarray`, if existing.
#         """
#         return (
#             self._numpy_ndarray.dtype
#             if self._numpy_ndarray is not None else
#             None
#         )
    
#     @property
#     def np_shape(
#         self
#     ) -> Union[tuple, None]:
#         """
#         The `shape` of the frame as a `np.ndarray`, if existing.
#         """
#         return (
#             self._numpy_ndarray.shape
#             if self._numpy_ndarray is not None else
#             None
#         )
    
#     # Attributes if 'moderngl.Texture'
#     @property
#     def gl_dtype(
#         self
#     ) -> Union[str, None]:
#         """
#         The `dtype` of the frame as a `moderngl.Texture`, if
#         existing.
#         """
#         return (
#             self._moderngl_texture.dtype
#             if self._moderngl_texture is not None else
#             None
#         )
    
#     @property
#     def gl_components(
#         self
#     ) -> Union[str, None]:
#         """
#         The number of `components` of the frame as a 
#         `moderngl.Texture`, if existing.
#         """
#         return (
#             self._moderngl_texture.components
#             if self._moderngl_texture is not None else
#             None
#         )

#     source_domain: FrameBufferDomain = field(default = None, init = False)
#     """
#     The domain that has been used as the source of
#     the frame, which means that is the one that was
#     used to create the instance and the rest, if
#     existing, have been created from it.

#     This source domain will also set when we modify
#     the frame and we write on it, so it becomes the
#     new source.
#     """

#     _av_videoframe: Union['av.VideoFrame', None] = field(default = None, init = False)
#     """
#     *For internal use only*

#     The frame as a `av.VideoFrame`.
#     """
#     _numpy_ndarray: Union['np.ndarray', None] = field(default = None, init = False)
#     """
#     *For internal use only*

#     The frame as a `numpy.ndarray`.
#     """
#     _moderngl_texture: Union['moderngl.Texture', None] = field(default = None, init = False)
#     """
#     *For internal use only*

#     The frame as a `moderngl.Texture`.
#     """

#     @property
#     def width(
#         self
#     ) -> int:
#         """
#         The width of the frame.
#         """
#         return self.size[0]
    
#     @property
#     def height(
#         self
#     ) -> int:
#         """
#         The height of the frame.
#         """
#         return self.size[1]
    
#     @property
#     def domains(
#         self
#     ) -> list[FrameBufferDomain]:
#         """
#         Get a list containing the domains available based
#         on the information we have in the different
#         domains (as numpy, texture or pyav frame).
#         """
#         domains = []

#         if self.has_gpu:
#             domains.append(FrameBufferDomain.GPU)

#         if self.has_cpu:
#             domains.append(FrameBufferDomain.CPU)

#         if self.has_media:
#             domains.append(FrameBufferDomain.MEDIA)

#         return domains
    
#     @property
#     def has_gpu(
#         self
#     ) -> bool:
#         """
#         Boolean flag to indicate if there is GPU
#         information stored, which means that the
#         `self._moderngl_texture` is not `None`.
#         """
#         return self._moderngl_texture is not None
    
#     @property
#     def has_cpu(
#         self
#     ) -> bool:
#         """
#         Boolean flag to indicate if there is CPU
#         information stored, which means that the
#         `self._numpy_ndarray` is not `None`.
#         """
#         return self._numpy_ndarray is not None
    
#     @property
#     def has_media(
#         self
#     ) -> bool:
#         """
#         Boolean flag to indicate if there is Media
#         information stored, which means that the
#         `self._av_videoframe` is not `None`.
#         """
#         return self._av_videoframe is not None
    
#     @classmethod
#     def from_frame(
#         cls,
#         frame: Union['av.VideoFrame', 'np.ndarray', 'moderngl.Texture', 'FrameBuffer']
#     ) -> 'FrameBuffer':
#         """
#         Instantiate a `FrameBuffer` from one of the following
#         frame types:
#         - `av.VideoFrame`
#         - `numpy.ndarray`
#         - `moderngl.Texture`
#         - `FrameBuffer`
#         """
#         ParameterValidator.validate_mandatory_instance_of('frame', frame, ['VideoFrame', 'ndarray', 'Texture', 'FrameBuffer'])

#         return (
#             FrameBuffer.from_av_videoframe(frame)
#             if PythonValidator.is_instance_of(frame, 'VideoFrame') else
#             FrameBuffer.from_numpy_ndarray(frame)
#             if PythonValidator.is_numpy_array(frame) else
#             FrameBuffer.from_moderngl_texture(frame)
#             if PythonValidator.is_instance_of(frame, 'Texture') else
#             frame
#         )
    
#     def _autodetect_transparency(
#         self
#     ) -> 'FrameBuffer':
#         """
#         *For internal use only*

#         Internal method to detect the transparency from
#         the frame that is available, using this priority:
#         - `cpu (np.ndarray)`
#         - `media (av.VideoFrame)`
#         - `gpu (moderngl.Texture)`
#         """
#         if self.has_cpu:
#             if (
#                 self._numpy_ndarray.ndim == 3 and
#                 self._numpy_ndarray.shape[2] == 4
#             ):
#                 alpha = self._numpy_ndarray[..., 3]
#                 a_min = alpha.min()
#                 a_max = alpha.max()
#                 self.has_alpha_pixels = a_min < 255
#                 self.is_fully_opaque = a_min == 255
#                 self.is_fully_transparent = a_max == 0
#             else:
#                 self.has_alpha_pixels = False
#                 self.is_fully_opaque = True
#                 self.is_fully_transparent = False
#         elif self.has_media:
#             """
#             TODO: We could force the '.reformat("rgba")' and store
#             the corresponding numpy array because we will need it
#             after to be processed (as CPU or to transform it to GPU).
#             """
#             self.has_alpha_pixels = 'a' in self._av_videoframe.format.name
#             self.is_fully_opaque = (
#                 True
#                 if not self.has_alpha_pixels else
#                 # We don't know yet
#                 None
#             )
#             self.is_fully_transparent = (
#                 False
#                 if not self.has_alpha_pixels else
#                 # We don't know yet
#                 None
#             )
#         elif self.has_gpu:
#             # TODO: By now we don't know it...
#             self.has_alpha_pixels = None
#             self.is_fully_opaque = None
#             self.is_fully_transparent = None

#         return self

#     def _autocalculate_numpy_from_av(
#         self,
#         frame: 'av.VideoFrame'
#     ) -> 'FrameBuffer':
#         """
#         *For internal use only*

#         Autocalculate the `np.ndarray` from the `frame`
#         given that must be an `av.VideoFrame`. We
#         pre-calculate it because we will need it soon.
#         """
#         self._numpy_ndarray = frame.to_ndarray(format = FORMAT_TO_FORCE)
#         # Detect the real transparency from the numpy
#         self._autodetect_transparency()

#         return self

#     @classmethod
#     @requires_dependency('av', 'yta_editor_frame', 'av')
#     def from_av_videoframe(
#         cls,
#         frame: 'av.VideoFrame'
#     ) -> 'FrameBuffer':
#         """
#         Instantiate a `FrameBuffer` from a `av.VideoFrame`.

#         This method will store the original frame as it is,
#         modifying not the type nor the format.
#         """
#         ParameterValidator.validate_mandatory_instance_of('frame', frame, 'VideoFrame')

#         # TODO: We could reformat and transform to 'ndarray'
#         # and read. We will need the 'ndarray' when processing
#         frame_buffer = cls(
#             size = (frame.width, frame.height)
#         )
#         frame_buffer._av_videoframe = frame
#         frame_buffer.source_domain = FrameBufferDomain.MEDIA

#         # Calculate and store the 'np.ndarray' because we
#         # will need it later
#         frame_buffer._autocalculate_numpy_from_av(frame)

#         return frame_buffer
    
#     @classmethod
#     @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
#     def from_numpy_ndarray(
#         cls,
#         frame: 'np.ndarray'
#     ) -> 'FrameBuffer':
#         """
#         Instantiate a `FrameBuffer` from a `numpy.ndarray`.

#         This method will store the original frame as it is,
#         modifying not the type nor the format.
#         """
#         ParameterValidator.validate_mandatory_numpy_array('frame', frame)

#         height, width = frame.shape[:2]

#         frame_buffer = cls(
#             size = (width, height)
#         )
#         frame_buffer._numpy_ndarray = frame
#         frame_buffer.source_domain = FrameBufferDomain.CPU
#         frame_buffer._autodetect_transparency()

#         return frame_buffer

#     @classmethod
#     @requires_dependency('moderngl', 'yta_editor_frame', 'moderngl')
#     def from_moderngl_texture(
#         cls,
#         frame: 'moderngl.Texture'
#     ) -> 'FrameBuffer':
#         """
#         Instantiate a `FrameBuffer` from a `moderngl.Texture`.
        
#         This method will store the original frame as it is,
#         modifying not the type nor the format.
#         """
#         ParameterValidator.validate_mandatory_instance_of('frame', frame, 'Texture')

#         width, height = frame.size

#         frame_buffer = cls(
#             size = (width, height),
#         )
#         frame_buffer._moderngl_texture = frame
#         frame_buffer.source_domain = FrameBufferDomain.GPU
#         frame_buffer._autodetect_transparency()

#         return frame_buffer
    
#     def _validate_source_available(
#         self
#     ):
#         """
#         Validate that there is at least one source to be
#         used to transform into the one we need for the
#         method that calls this validation.

#         This method will raise an exception if there is
#         not a single domain available.
#         """
#         if len(self.domains) == 0:
#             raise RuntimeError('No source available in this FrameBuffer instance.')
        
#     """
#     TODO: Maybe I need the 'requires_dependency' in some
#     of these 'as_...' methods
#     """

#     def update_frame(
#         self,
#         frame: Union['np.ndarray', 'moderngl.Texture', 'av.VideoFrame']
#     ) -> 'FrameBuffer':
#         """
#         Update the internal frame, setting it as the only
#         source and removing the previous ones.
#         """
#         ParameterValidator.validate_mandatory_instance_of('frame', frame, ['ndarray', 'Texture', 'VideoFrame'])

#         if PythonValidator.is_instance_of(frame, 'ndarray'):
#             self._numpy_ndarray = frame
#             self._av_videoframe = None
#             self._moderngl_texture = None
#         elif PythonValidator.is_instance_of(frame, 'Texture'):
#             self._numpy_ndarray = None
#             self._av_videoframe = None
#             self._moderngl_texture = frame
#         elif PythonValidator.is_instance_of(frame, 'VideoFrame'):
#             self._numpy_ndarray = None
#             self._av_videoframe = frame
#             self._moderngl_texture = None

#             self._autocalculate_numpy_from_av(frame)
        
#         self._autodetect_transparency()
    
#     @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
#     def as_numpy_ndarray(
#         self,
#         do_flip_texture: bool = True,
#         do_force_contiguous: bool = False
#     ) -> 'np.ndarray':
#         """
#         Get the frame but as a `numpy.ndarray`. This is
#         meant for CPU use.

#         This method will lazy load the value from the
#         one that is available in the moment it is
#         requested.

#         This method will always force the `rgba` format.
#         """
#         import numpy as np

#         self._validate_source_available()

#         if self._numpy_ndarray is None:
#             if self._av_videoframe is not None:
#                 # 1. `av.VideoFrame` -> `np.ndarray`
#                 # 2. Force 'rgba' format
#                 frame = self._av_videoframe.to_ndarray(format = FORMAT_TO_FORCE)

#                 # Flipping it would be unexpected, so we don't
#                 frame = (
#                     np.ascontiguousarray(frame)
#                     if do_force_contiguous else
#                     frame
#                 )

#                 self._numpy_ndarray = frame
#                 self._autodetect_transparency()
#             elif self._moderngl_texture is not None:
#                 # 1. `moderngl.Texture` -> `np.ndarray`
#                 frame = _texture_to_numpy_frame(
#                     texture = self._moderngl_texture,
#                     size = (self.width, self.height),
#                     do_flip_texture = do_flip_texture,
#                     do_force_contiguous = False
#                 )

#                 # 2. Force 'rgba' format
#                 frame = _ensure_rgba_numpy_array(
#                     ndarray = frame,
#                     do_force_contiguous = do_force_contiguous
#                 )

#                 self._numpy_ndarray = frame
#                 self._autodetect_transparency()

#         return self._numpy_ndarray
    
#     def as_moderngl_texture(
#         self,
#         moderngl_context: 'moderngl.Context',
#         do_flip_texture: bool = True
#     ) -> 'moderngl.Texture':
#         """
#         Get the frame but as a `moderngl.Texture`. This
#         is meant for GPU use.

#         This method will lazy load the value from the
#         one that is available in the moment it is
#         requested.

#         Remember that when obtaining this parameter
#         """
#         self._validate_source_available()

#         if self._moderngl_texture is None:
#             if self._numpy_ndarray is not None:
#                 # 1. Force 'rgba' format
#                 frame = _ensure_rgba_numpy_array(
#                     ndarray = self._numpy_ndarray,
#                     do_force_contiguous = False
#                 )

#                 # 2. `np.ndarray` -> `moderngl.Texture`
#                 self._moderngl_texture = _numpy_frame_to_texture(
#                     numpy_frame = frame,
#                     moderngl_context = moderngl_context,
#                     size = (self.width, self.height),
#                     do_flip_texture = do_flip_texture
#                 )

#                 # We preserve the information about the alpha
#                 # channel that we had when numpy array
#             elif self._av_videoframe is not None:
#                 # 1. `av.VideoFrame` -> `np.ndarray`
#                 # 2. Force 'rgba' format
#                 frame = self._av_videoframe.to_ndarray(format = FORMAT_TO_FORCE)

#                 # As we obtained it as numpy array, preserve it and
#                 # the information about the alpha channel from it
#                 self._numpy_ndarray = frame
#                 self._autodetect_transparency()

#                 # 3. `np.ndarray` -> `moderngl.Texture`
#                 self._moderngl_texture = _numpy_frame_to_texture(
#                     numpy_frame = frame,
#                     moderngl_context = moderngl_context,
#                     size = (self.width, self.height),
#                     do_flip_texture = do_flip_texture
#                 )
                
#         return self._moderngl_texture
    
#     @requires_dependency('av', 'yta_editor_frame', 'av')
#     def as_av_videoframe(
#         self,
#         do_flip_texture: bool = True
#     ) -> 'av.VideoFrame':
#         """
#         Get the frame but as a `av.VideoFrame`.

#         This method will lazy load the value from the
#         one that is available in the moment it is
#         requested.
#         """
#         import av

#         self._validate_source_available()

#         if self._av_videoframe is None:
#             if self._numpy_ndarray is not None:
#                 # 1. Force 'rgba' format
#                 frame = _ensure_rgba_numpy_array(
#                     ndarray = self._numpy_ndarray,
#                     do_force_contiguous = True
#                 )

#                 # 2. `np.ndarray` -> `av.VideoFrame`
#                 self._av_videoframe = av.VideoFrame.from_ndarray(frame, format = FORMAT_TO_FORCE)

#                 # We preserve the information about the alpha
#                 # channel that we had when numpy array
#             elif self._moderngl_texture is not None:
#                 # 1. `moderngl.Texture` -> `np.ndarray`
#                 frame = _texture_to_numpy_frame(
#                     texture = self._moderngl_texture,
#                     size = (self.width, self.height),
#                     do_flip_texture = do_flip_texture,
#                     do_force_contiguous = True
#                 )

#                 # As we obtained it as numpy array, preserve it and
#                 # the information about the alpha channel from it
#                 self._numpy_ndarray = frame
#                 self._autodetect_transparency()

#                 # 2. `np.ndarray` -> `av.VideoFrame`
#                 self._av_videoframe = av.VideoFrame.from_ndarray(frame, format = FORMAT_TO_FORCE)

#         return self._av_videoframe


# """
# A frame in HWC format (most common) is like this:
# - `h, w = frame.shape[:2]`

# If grayscale, is like this:
# - `h, w = frame.shape`

# If CHW, is like this:
# - `c, h, w = frame.shape`
# """

# """
# Puntos críticos (muy importantes en tu pipeline):

# Flip vertical (OpenGL vs imagen)
# OpenGL tiene el origen en bottom-left.
# NumPy / PyAV esperan top-left.

# Debes invertir verticalmente:
# - `frame = np.flipud(frame)`

# Número de canales:
# - `texture.components`

# 3 → RGB
# 4 → RGBA

# Tipo de dato
# Por defecto:
# - `dtype = uint8`

# Si usas texturas float:
# - `frame = np.frombuffer(data, dtype = np.float32)`

# Contigüidad
# Después del flipud, fuerza contigüidad:
# - `frame = np.ascontiguousarray(frame)`
# """


from abc import abstractmethod, ABC


class FrameResourceABC(ABC):
    """
    Abstract method to represent a frame resource.
    """

    @property
    @abstractmethod
    def width(
        self
    ) -> int:
        """
        The width of the frame resource, in pixels.
        """
        pass

    @property
    @abstractmethod
    def height(
        self
    ) -> int:
        """
        The height of the frame resource, in pixels.
        """
        pass

    @property
    @abstractmethod
    def format(
        self
    ) -> int:
        """
        The format of the frame resource.

        TODO: Is this useful (?)
        """
        pass

    @property
    @abstractmethod
    def type(
        self
    ) -> 'FrameResourceType':
        """
        The type of frame resource.
        """
        pass

    @abstractmethod
    def release(
        self,
        render_context: 'RenderContext'
    ):
        """
        Release the resource, if needed.
        """
        raise NotImplementedError


from yta_editor_frame.frame import Frame
from yta_validation import PythonValidator
from yta_validation.parameter import ParameterValidator
from typing import Union
from functools import wraps
from collections.abc import Mapping


def force_frame(
    *arg_names
):
    """
    Decorator to transform the parameter provided
    into `Frame` instances, if the parameters are
    one of the following types:
    - `av.VideoFrame`
    - `np.ndarray`
    - `moderngl.Texture`
    - `RenderTarget`

    A `RenderContext` instance is needed as
    `render_context` to be able to make the
    transformation.

    This method could need one of the following optional
    libraries
    to work:
    - `av`
    - `moderngl`
    - `numpy`

    The code must be like this below:
    ```
    @force_frame('inputs')
    def test(
        inputs: dict[str, any]
    ):
        # Here the 'inputs' is now a dict with 'Frame'
        pass
    ```
    """
    def decorator(
        func
    ):
        @wraps(func)
        def wrapper(
            *args,
            **kwargs
        ):
            def convert_to_frame(
                input: Union['Frame', 'RenderTarget', 'moderngl.Texture', 'np.ndarray', 'av.VideoFrame', Mapping, None],
                render_context: 'RenderContext'
            ):
                """
                Transform all the inputs provided in the `input`
                parameter to a `Frame` instance. The `input`
                parameter can be a single input or a dict of them.
                """
                # We can have something like 'base_input': None
                # that we accept for some specific nodes
                if (
                    input is None or
                    PythonValidator.is_instance_of(input, 'Frame')
                ):
                    return input

                if not PythonValidator.is_instance_of(input, ['Frame', 'RenderTarget', 'Texture', 'ndarray', 'VideoFrame', Mapping]):
                    raise Exception(f'We cannot convert "{type(input)}" to Frame.')
                
                return (
                    {
                        key: convert_to_frame(
                            input = input_item,
                            render_context = render_context
                        )
                        for key, input_item in input.items()
                    }
                    if PythonValidator.is_instance_of(input, Mapping) else
                    Frame.from_frame(
                        frame = input,
                        render_context = render_context
                    )
                )
            
            # args[0] is self
            if (
                len(args) == 0 or
                args[0]._render_context is None
            ):
                raise Exception('The "._render_context" attribute was not found.')
            
            render_context = args[0]._render_context

            ParameterValidator.validate_mandatory_instance_of('render_context', render_context, 'RenderContext')

            for name in arg_names:
                if name in kwargs:
                    kwargs[name] = convert_to_frame(
                        input = kwargs[name],
                        render_context = render_context
                    )

            return func(*args, **kwargs)
        return wrapper
    return decorator


from yta_editor_frame.frame.resource.types import FrameResourceType


class FrameResourceIdentifier:
    """
    Class to simplify the way we identify the type
    of `FrameResource` instances we are dealing
    with.
    """

    @staticmethod
    def is_cpu(
        frame_resource: 'FrameResource'
    ) -> bool:
        try:
            return frame_resource.type == FrameResourceType.CPU
        except:
            return False
        
    @staticmethod
    def is_gpu(
        frame_resource: 'FrameResource'
    ) -> bool:
        try:
            return frame_resource.type == FrameResourceType.GPU
        except:
            return False
        
    @staticmethod
    def is_cpu_av_videoframe(
        frame_resource: 'FrameResource'
    ) -> bool:
        try:
            return frame_resource.type == FrameResourceType.CPU_AV_VIDEOFRAME
        except:
            return False
    
    @staticmethod
    def is_gpu_render_target(
        frame_resource: 'FrameResource'
    ) -> bool:
        try:
            return frame_resource.type == FrameResourceType.GPU_RENDER_TARGET
        except:
            return False


"""
I don't think we need this...
"""
from yta_validation import PythonValidator
from dataclasses import dataclass
from typing import Union

import numpy as np


@dataclass(frozen = True)
class FrameFormatDescriptor:
    """
    Dataclass to identify and simplify the way we
    handle the different formats.
    """

    name: str
    """
    The name of the format.
    """
    number_of_channels: int
    """
    The amount of channels included.
    """
    numpy_dtype: np.dtype
    """
    The raw numpy 'dtype' format.
    """
    av_format: str
    """
    The name of the format for the PyAV library.
    """
    gl_components: Union[int, None]
    """
    The number of components when as a moderngl
    texture, that can be `None` if it needs to be
    converted.
    """


FRAME_FORMATS = {
    'rgba8': FrameFormatDescriptor(
        name = 'rgba8',
        number_of_channels = 4,
        numpy_dtype = np.uint8,
        av_format = 'rgba',
        gl_components = 4,
    ),
    'rgb8': FrameFormatDescriptor(
        name = 'rgb8',
        number_of_channels = 3,
        numpy_dtype = np.uint8,
        av_format = 'rgb24',
        gl_components = 3,
    ),
    'gray8': FrameFormatDescriptor(
        name = 'gray8',
        number_of_channels = 1,
        numpy_dtype = np.uint8,
        av_format = 'gray',
        gl_components = 1,
    ),
    'yuv420p': FrameFormatDescriptor(
        name = 'yuv420p',
        number_of_channels = 3,
        numpy_dtype = np.uint8,
        av_format = 'yuv420p',
        gl_components = None,  # Need to transform to upload
    ),
}

class FrameFormatDetector:
    """
    Class to simplify the way we detect the format
    of a video frame.
    """

    # TODO: Set the 'av' library requirement
    @staticmethod
    def _from_pyav_frame(
        frame: 'av.VideoFrame'
    ) -> FrameFormatDescriptor:
        """
        *For internal use only*

        Detect the format descriptor from the pyav
        video frame `frame` provided.

        This will detect it by checking the format name
        of the `frame` provided.
        """
        import av

        # TODO: Validate the 'frame' is a 'av.VideoFrame' (?)
        format = frame.format.name

        for frame_format_descriptor in FRAME_FORMATS.values():
            if frame_format_descriptor.av_format == format:
                return frame_format_descriptor

        raise Exception(f'The "{format}" is not supported.')
    
    # TODO: Set the 'numpy' library requirement
    @staticmethod
    def _from_numpy_frame(
        frame: 'np.ndarray'
    ) -> FrameFormatDescriptor:
        """
        *For internal use only*

        Detect the format descriptor from the numpy
        video frame `frame` provided.

        This will detect it by checking the number of
        `channels` and the `dtype` in the `frame`
        provided.
        """
        import numpy as np

        # TODO: Validate the 'frame' is a 'np.ndarray' frame (?)
        number_of_channels = (
            1
            if frame.ndim == 2 else
            frame.shape[2]
        )
        dtype = np.dtype(frame.dtype)

        for frame_format_descriptor in FRAME_FORMATS.values():
            if (
                frame_format_descriptor.number_of_channels == number_of_channels and
                frame_format_descriptor.numpy_dtype == dtype
            ):
                return frame_format_descriptor
            
        raise Exception(f'The numpy ndarray with {str(number_of_channels)} channels and dtype={str(dtype)} is not supported.')
    
    # TODO: Set the 'moderngl' library requirement
    @staticmethod
    def _from_moderngl_frame(
        frame: 'moderngl.Texture'
    ) -> FrameFormatDescriptor:
        """
        *For internal use only*

        Detect the format descriptor from the moderngl
        video frame `frame` provided.

        This will detect it by checking the number of
        `components` existing in the `frame` provided.
        """
        import moderngl

        number_of_components = frame.components

        for desc in FRAME_FORMATS.values():
            if desc.gl_components == number_of_components:
                return desc

        raise Exception(f'The moderngl texture with {str(number_of_components)} components is not supported')
    
    # TODO: Set the 'av' library requirement
    # TODO: Set the 'numpy' library requirement
    # TODO: Set the 'moderngl' library requirement
    @staticmethod
    def from_frame(
        frame: Union['np.ndarray', 'av.VideoFrame', 'moderngl.Texture']
    ) -> Union[FrameFormatDescriptor, None]:
        """
        Detect the format descriptor from the video
        frame `frame` provided, that can be one of
        these:
        - `np.ndarray`
        - `av.VideoFrame`
        - `moderngl.Texture`
        """
        return (
            FrameFormatDetector._from_pyav_frame(frame)
            if PythonValidator.is_instance_of(frame, 'VideoFrame') else
            FrameFormatDetector._from_numpy_frame(frame)
            if PythonValidator.is_numpy_array(frame) else
            FrameFormatDetector._from_moderngl_frame(frame)
            if PythonValidator.is_instance_of(frame, 'Texture') else
            None
        )


from yta_editor_frame.frame.resource.abstract import FrameResourceABC
from yta_editor_frame.frame.resource.types import FrameResourceType


class GPUTextureFrameResource(
    FrameResourceABC
):
    """
    A frame resource based on a texture.

    The `.frame` attribute is a `moderngl.Texture`.
    """
    
    @property
    def width(
        self
    ) -> int:
        return self.frame.width
    
    @property
    def height(
        self
    ) -> int:
        return self.frame.height
    
    @property
    def format(
        self
    ) -> str:
        return self.frame.dtype
    
    @property
    def type(
        self
    ) -> FrameResourceType:
        return FrameResourceType.GPU

    def __init__(
        self,
        frame: 'moderngl.Texture'
    ):
        self.frame: 'moderngl.Texture' = frame
        """
        The `frame` itself, as a `moderngl.Texture`.
        """

    def release(
        self,
        render_context: 'RenderContext'
    ):
        render_context.render_resources.texture.release(self.frame)
        


"""
This module is recent, used with `FrameBuffer`
and working properly, a bit more strict than
others you can find in the project and maybe
should be removed.
"""
from yta_programming.decorators.requires_dependency import requires_dependency
from yta_editor_pools.texture.provider import TextureProvider


@requires_dependency('numpy', 'yta_editor_frame', 'numpy')
def _texture_to_numpy_frame(
    texture: 'moderngl.Texture',
    size: tuple[int, int],
    do_flip_texture: bool = True,
    do_force_contiguous: bool = False
):
    """
    *For internal use only*

    Create a `np.ndarray` from the `texture` provided
    and return it.

    This method is only meant to be used within
    the `FrameBuffer` class.

    This frame is not forced to be contiguous.
    """
    import numpy as np

    data = texture.read()
    """
    We are always using 'np.uint8' because we are also
    forcing 'f1' textures dtype.
    """
    frame = np.frombuffer(data, dtype = np.uint8)
    # The format is (height, width, number_of_components)
    frame = frame.reshape(size[1], size[0], texture.components)

    # We need to flip it depending on our moderngl shaders
    frame = (
        np.flipud(frame)
        if do_flip_texture else
        frame
    )

    # We force it as contiguous
    frame = (
        np.ascontiguousarray(frame)
        if do_force_contiguous else
        frame
    )

    # # We always force 'RGBA'
    # frame = _ensure_rgba_numpy_array(frame)

    return frame

@requires_dependency('numpy', 'yta_editor_frame', 'numpy')
def _ensure_rgba_numpy_array(
    ndarray: 'np.ndarray',
    do_force_contiguous: bool = False
) -> 'np.ndarray':
    """
    *For internal use only*

    Transform the `ndarray` provided to make sure
    that it is representing a 'rgba' frame,
    compatible with textures like the following one:
    - `dtype = uint8`
    - `shape = (H, W, 4)`
    - `layout = RGBA`

    This method is only meant to be used within
    the `FrameBuffer` class.

    This array is forced to be contiguous if the
    `do_force_contiguous` parameter is `True`, that
    must be set as `True` when we are about to use
    that frame to build a texture o to write with
    the pyav library.
    """
    import numpy as np

    if ndarray.ndim == 2:
        ndarray = np.repeat(ndarray[:, :, None], 3, axis = 2)

    if ndarray.shape[2] == 3:
        alpha = np.full((ndarray.shape[0], ndarray.shape[1], 1), 255, dtype = ndarray.dtype)
        ndarray = np.concatenate((ndarray, alpha), axis = 2)

    if ndarray.shape[2] != 4:
        raise Exception('The format of this numpy array is not accepted.')
    
    ndarray = (
        np.ascontiguousarray(ndarray)
        if do_force_contiguous else
        ndarray
    )

    return ndarray

@requires_dependency('numpy', 'yta_editor_frame', 'numpy')
@requires_dependency('moderngl', 'yta_editor_frame', 'moderngl')
def _numpy_frame_to_texture(
    numpy_frame: 'np.ndarray',
    moderngl_context: 'moderngl.Context',
    size: tuple[int, int],
    do_flip_texture: bool = True
) -> 'moderngl.Texture':
    """
    *For internal use only*

    Create a `moderngl.Texture` from the given
    `numpy_frame` and return it.

    This method is only meant to be used within
    the `FrameBuffer` class.

    Please, remember to release the texture when
    finished as it is being extracted from the
    pool.
    """
    import moderngl
    import numpy as np

    numpy_frame = (
        np.flipud(numpy_frame)
        if do_flip_texture else
        numpy_frame
    )

    number_of_components = (
        numpy_frame.shape[2]
        if numpy_frame.ndim == 3 else
        1
    )

    # We force contiguous for moderngl
    numpy_frame = np.ascontiguousarray(numpy_frame)

    """
    We are always using 'np.uint8' because we are also
    forcing 'f1' textures dtype.
    """
    # TODO: Sometimes we can receive a 'float32', why (?)
    # TODO: Transform dtype to texture dtype with utils
    dtype = (
        'f1'
        if numpy_frame.dtype.name == 'uint8' else
        'f4'
        if numpy_frame.dtype.name == 'float32' else
        # This is not ok, but by now...
        'f1'
    )

    # We need to release it this one when finished
    texture = TextureProvider(moderngl_context).acquire(
        width = size[0],
        height = size[1],
        number_of_components = number_of_components,
        dtype = dtype
    )
    texture.write(numpy_frame)

    # texture = moderngl_context.texture(
    #     size = size,
    #     components = number_of_components,
    #     data = numpy_frame.tobytes(),
    #     dtype = dtype
    # )

    # Default filter
    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)

    return texture


"""
Module to transform one resource into another
one to be able to do CPU -> GPU, GPU -> CPU,
etc.
"""
from yta_editor_frame.frame.resource.identifier import FrameResourceIdentifier
from yta_editor_frame.frame.resource.cpu_frame_resource import CPUNumpyArrayFrameResource
from yta_editor_frame.frame.resource.gpu_frame_resource import GPUTextureFrameResource
from yta_editor_frame.frame.resource.av_video_frame_resource import AvVideoFrameResource
from yta_programming.decorators.requires_dependency import requires_dependency
from typing import Union


FORMAT = 'rgba'

class FrameResourceConverter:
    """
    Class to simplify the way we transform one
    frame resource into another one.
    """

    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def to_numpy(
        resource: Union[CPUNumpyArrayFrameResource, GPUTextureFrameResource]
    ) -> CPUNumpyArrayFrameResource:
        """
        Transform the `resource` provided into a GPU
        `CPUNumpyArrayFrameResource` that holds a
        `np.ndarray`.
        """
        if FrameResourceIdentifier.is_cpu(resource):
            # [1] `np.ndarray` -> `np.ndarray`
            return resource
        
        frame = None
        if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
            # [2] `av.VideoFrame` -> `np.ndarray`
            import numpy as np

            # TODO: av to numpy
            numpy_array = resource.frame.to_ndarray(format = FORMAT)

            # Flipping it would be unexpected, so we don't
            # frame = (
            #     np.ascontiguousarray(frame)
            #     if do_force_contiguous else
            #     frame
            # )

            # self._numpy_ndarray = frame
            # self._autodetect_transparency()
            frame = numpy_array
        else:
            # [3] `moderngl.Texture` -> `np.ndarray`
            # [4] `render_target` -> `np.ndarray`
            from yta_editor_frame.utils import _texture_to_numpy_frame, _ensure_rgba_numpy_array

            # [3] `moderngl.Texture` -> `np.ndarray`
            texture = resource.frame
            if FrameResourceIdentifier.is_gpu_render_target(resource):
                # [4] `render_target` -> `np.ndarray`
                # TODO: What do we do with the render target? Should
                # this conversion happen? I think it shouldn't...
                texture = texture.texture

            # `moderngl.Texture` -> `np.ndarray`
            numpy_frame = _texture_to_numpy_frame(
                texture = texture,
                size = (resource.width, resource.height),
                # TODO: What do we do with the flip (?)
                do_flip_texture = True,
                # do_flip_texture = do_flip_texture,
                do_force_contiguous = False
            )

            # 2. Force 'rgba' format
            frame = _ensure_rgba_numpy_array(
                ndarray = numpy_frame,
                # TODO: What do we do with the flip (?)
                # do_force_contiguous = do_force_contiguous
                do_force_contiguous = False
            )

            # self._numpy_ndarray = frame
            # self._autodetect_transparency()
            # TODO: Do we need to release the `render_target`
            # or `moderngl.Texture` to the pool (?)

        return CPUNumpyArrayFrameResource(frame)
        
    @requires_dependency('numpy', 'yta_editor_frame', 'numpy')
    def to_texture(
        resource: 'FrameResource',
        moderngl_context: 'moderngl.Context'
    ) -> GPUTextureFrameResource:
        """
        Transform the `resource` provided into a GPU
        `GPUTextureFrameResource` that holds a
        `moderngl.Texture`.
        """
        if FrameResourceIdentifier.is_gpu(resource):
            # [1] `moderngl.Texture` -> `moderngl.Texture`
            return resource

        frame = None
        if FrameResourceIdentifier.is_gpu_render_target(resource):
            # [2] `render_target` -> `moderngl.Texture`
            # TODO: What do we do with the render target? Should
            # this conversion happen? I think it shouldn't...
            frame = resource.frame.texture
        elif (
            FrameResourceIdentifier.is_cpu(resource) or
            FrameResourceIdentifier.is_cpu_av_videoframe(resource)
        ):
            from yta_editor_frame.utils import _ensure_rgba_numpy_array, _numpy_frame_to_texture
            # [3] `np.ndarray` -> `moderngl.Texture`
            # [4] `av.VideoFrame` -> `moderngl.Texture`

            numpy_array = resource.frame
            if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
                # [4] `av.VideoFrame` -> `moderngl.Texture`
                numpy_array = numpy_array.to_ndarray(format = FORMAT)

            # [3] `np.ndarray` -> `moderngl.Texture`
            # 1. Force 'rgba' format
            numpy_array = _ensure_rgba_numpy_array(
                ndarray = numpy_array,
                do_force_contiguous = False
            )

            # 2. `np.ndarray` -> `moderngl.Texture`
            frame = _numpy_frame_to_texture(
                numpy_frame = numpy_array,
                moderngl_context = moderngl_context,
                size = (resource.width, resource.height),
                # TODO: What do we do with the flip (?)
                do_flip_texture = True
                # do_flip_texture = do_flip_texture
            )

            # We preserve the information about the alpha
            # channel that we had when numpy array

        return GPUTextureFrameResource(frame)
    
    @requires_dependency('av', 'yta_editor_frame', 'av')
    def to_videoframe(
        resource: 'FrameResource'
    ) -> AvVideoFrameResource:
        """
        Transform the `resource` provided into a GPU
        `AvVideoFrameResource` that holds a
        `av.VideoFrame`.
        """
        import av
        
        if FrameResourceIdentifier.is_cpu_av_videoframe(resource):
            # [1] `av.VideoFrame` -> `av.VideoFrame`
            return resource
        
        frame = None
        if FrameResourceIdentifier.is_cpu(resource):
            # [2] `np.ndarray` -> `av.VideoFrame`
            from yta_editor_frame.utils import _ensure_rgba_numpy_array
            # 1. Force 'rgba' format
            numpy_array = _ensure_rgba_numpy_array(
                ndarray = resource.frame,
                do_force_contiguous = True
            )

            # 2. `np.ndarray` -> `av.VideoFrame`
            frame = av.VideoFrame.from_ndarray(numpy_array, format = FORMAT)

            # We preserve the information about the alpha
            # channel that we had when numpy array
        else:
            # [3] `moderngl.Texture` -> `av.VideoFrame`
            # [4] `render_target` -> `av.VideoFrame`
            from yta_editor_frame.utils import _texture_to_numpy_frame

            frame = resource.frame
            if FrameResourceIdentifier.is_gpu_render_target(resource):
                # [4] `render_target` -> `av.VideoFrame`
                frame = frame.texture

            # [3] `moderngl.Texture` -> `av.VideoFrame`
            # 1. `moderngl.Texture` -> `np.ndarray`
            numpy_array = _texture_to_numpy_frame(
                texture = frame,
                size = (resource.width, resource.height),
                # TODO: What do we do with the flip (?)
                do_flip_texture = True,
                # do_flip_texture = do_flip_texture,
                do_force_contiguous = True
            )

            # As we obtained it as numpy array, preserve it and
            # the information about the alpha channel from it
            # self._numpy_ndarray = frame
            # self._autodetect_transparency()

            # 2. `np.ndarray` -> `av.VideoFrame`
            frame = av.VideoFrame.from_ndarray(numpy_array, format = FORMAT)
            # TODO: Do we need to release the `render_target`
            # or `moderngl.Texture` to the pool (?)
        
        return AvVideoFrameResource(frame)
